# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: openfl/protocols/base.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1bopenfl/protocols/base.proto\"+\n\nModelProto\x12\x1d\n\x07tensors\x18\x01 \x03(\x0b\x32\x0c.NamedTensor\"\xa3\x01\n\x0bNamedTensor\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x14\n\x0cround_number\x18\x02 \x01(\x05\x12\x10\n\x08lossless\x18\x03 \x01(\x08\x12\x0e\n\x06report\x18\x04 \x01(\x08\x12\x0c\n\x04tags\x18\x05 \x03(\t\x12,\n\x14transformer_metadata\x18\x06 \x03(\x0b\x32\x0e.MetadataProto\x12\x12\n\ndata_bytes\x18\x07 \x01(\x0c\"\xac\x01\n\rMetadataProto\x12\x34\n\x0cint_to_float\x18\x01 \x03(\x0b\x32\x1e.MetadataProto.IntToFloatEntry\x12\x10\n\x08int_list\x18\x02 \x03(\x05\x12\x11\n\tbool_list\x18\x03 \x03(\x08\x12\r\n\x05model\x18\x04 \x01(\x08\x1a\x31\n\x0fIntToFloatEntry\x12\x0b\n\x03key\x18\x01 \x01(\x05\x12\r\n\x05value\x18\x02 \x01(\x02:\x02\x38\x01\"+\n\nDataStream\x12\x0c\n\x04size\x18\x01 \x01(\r\x12\x0f\n\x07npbytes\x18\x02 \x01(\x0c\"\x81\x01\n\x17\x43ollaboratorDescription\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0e\n\x06status\x18\x02 \x01(\t\x12\x10\n\x08progress\x18\x03 \x01(\x02\x12\r\n\x05round\x18\x04 \x01(\r\x12\x14\n\x0c\x63urrent_task\x18\x05 \x01(\t\x12\x11\n\tnext_task\x18\x06 \x01(\t\"4\n\x0fTaskDescription\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x02 \x01(\t\".\n\x0e\x44ownloadStatus\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0e\n\x06status\x18\x02 \x01(\t\"R\n\x10\x44ownloadStatuses\x12\x1f\n\x06models\x18\x01 \x03(\x0b\x32\x0f.DownloadStatus\x12\x1d\n\x04logs\x18\x02 \x03(\x0b\x32\x0f.DownloadStatus\"\xf4\x01\n\x15\x45xperimentDescription\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0e\n\x06status\x18\x02 \x01(\t\x12\x10\n\x08progress\x18\x03 \x01(\x02\x12\x14\n\x0ctotal_rounds\x18\x04 \x01(\r\x12\x15\n\rcurrent_round\x18\x05 \x01(\r\x12,\n\x11\x64ownload_statuses\x18\x06 \x01(\x0b\x32\x11.DownloadStatuses\x12/\n\rcollaborators\x18\x07 \x03(\x0b\x32\x18.CollaboratorDescription\x12\x1f\n\x05tasks\x18\x08 \x03(\x0b\x32\x10.TaskDescriptionb\x06proto3')



_MODELPROTO = DESCRIPTOR.message_types_by_name['ModelProto']
_NAMEDTENSOR = DESCRIPTOR.message_types_by_name['NamedTensor']
_METADATAPROTO = DESCRIPTOR.message_types_by_name['MetadataProto']
_METADATAPROTO_INTTOFLOATENTRY = _METADATAPROTO.nested_types_by_name['IntToFloatEntry']
_DATASTREAM = DESCRIPTOR.message_types_by_name['DataStream']
_COLLABORATORDESCRIPTION = DESCRIPTOR.message_types_by_name['CollaboratorDescription']
_TASKDESCRIPTION = DESCRIPTOR.message_types_by_name['TaskDescription']
_DOWNLOADSTATUS = DESCRIPTOR.message_types_by_name['DownloadStatus']
_DOWNLOADSTATUSES = DESCRIPTOR.message_types_by_name['DownloadStatuses']
_EXPERIMENTDESCRIPTION = DESCRIPTOR.message_types_by_name['ExperimentDescription']
ModelProto = _reflection.GeneratedProtocolMessageType('ModelProto', (_message.Message,), {
  'DESCRIPTOR' : _MODELPROTO,
  '__module__' : 'openfl.protocols.base_pb2'
  # @@protoc_insertion_point(class_scope:ModelProto)
  })
_sym_db.RegisterMessage(ModelProto)

NamedTensor = _reflection.GeneratedProtocolMessageType('NamedTensor', (_message.Message,), {
  'DESCRIPTOR' : _NAMEDTENSOR,
  '__module__' : 'openfl.protocols.base_pb2'
  # @@protoc_insertion_point(class_scope:NamedTensor)
  })
_sym_db.RegisterMessage(NamedTensor)

MetadataProto = _reflection.GeneratedProtocolMessageType('MetadataProto', (_message.Message,), {

  'IntToFloatEntry' : _reflection.GeneratedProtocolMessageType('IntToFloatEntry', (_message.Message,), {
    'DESCRIPTOR' : _METADATAPROTO_INTTOFLOATENTRY,
    '__module__' : 'openfl.protocols.base_pb2'
    # @@protoc_insertion_point(class_scope:MetadataProto.IntToFloatEntry)
    })
  ,
  'DESCRIPTOR' : _METADATAPROTO,
  '__module__' : 'openfl.protocols.base_pb2'
  # @@protoc_insertion_point(class_scope:MetadataProto)
  })
_sym_db.RegisterMessage(MetadataProto)
_sym_db.RegisterMessage(MetadataProto.IntToFloatEntry)

DataStream = _reflection.GeneratedProtocolMessageType('DataStream', (_message.Message,), {
  'DESCRIPTOR' : _DATASTREAM,
  '__module__' : 'openfl.protocols.base_pb2'
  # @@protoc_insertion_point(class_scope:DataStream)
  })
_sym_db.RegisterMessage(DataStream)

CollaboratorDescription = _reflection.GeneratedProtocolMessageType('CollaboratorDescription', (_message.Message,), {
  'DESCRIPTOR' : _COLLABORATORDESCRIPTION,
  '__module__' : 'openfl.protocols.base_pb2'
  # @@protoc_insertion_point(class_scope:CollaboratorDescription)
  })
_sym_db.RegisterMessage(CollaboratorDescription)

TaskDescription = _reflection.GeneratedProtocolMessageType('TaskDescription', (_message.Message,), {
  'DESCRIPTOR' : _TASKDESCRIPTION,
  '__module__' : 'openfl.protocols.base_pb2'
  # @@protoc_insertion_point(class_scope:TaskDescription)
  })
_sym_db.RegisterMessage(TaskDescription)

DownloadStatus = _reflection.GeneratedProtocolMessageType('DownloadStatus', (_message.Message,), {
  'DESCRIPTOR' : _DOWNLOADSTATUS,
  '__module__' : 'openfl.protocols.base_pb2'
  # @@protoc_insertion_point(class_scope:DownloadStatus)
  })
_sym_db.RegisterMessage(DownloadStatus)

DownloadStatuses = _reflection.GeneratedProtocolMessageType('DownloadStatuses', (_message.Message,), {
  'DESCRIPTOR' : _DOWNLOADSTATUSES,
  '__module__' : 'openfl.protocols.base_pb2'
  # @@protoc_insertion_point(class_scope:DownloadStatuses)
  })
_sym_db.RegisterMessage(DownloadStatuses)

ExperimentDescription = _reflection.GeneratedProtocolMessageType('ExperimentDescription', (_message.Message,), {
  'DESCRIPTOR' : _EXPERIMENTDESCRIPTION,
  '__module__' : 'openfl.protocols.base_pb2'
  # @@protoc_insertion_point(class_scope:ExperimentDescription)
  })
_sym_db.RegisterMessage(ExperimentDescription)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _METADATAPROTO_INTTOFLOATENTRY._options = None
  _METADATAPROTO_INTTOFLOATENTRY._serialized_options = b'8\001'
  _MODELPROTO._serialized_start=31
  _MODELPROTO._serialized_end=74
  _NAMEDTENSOR._serialized_start=77
  _NAMEDTENSOR._serialized_end=240
  _METADATAPROTO._serialized_start=243
  _METADATAPROTO._serialized_end=415
  _METADATAPROTO_INTTOFLOATENTRY._serialized_start=366
  _METADATAPROTO_INTTOFLOATENTRY._serialized_end=415
  _DATASTREAM._serialized_start=417
  _DATASTREAM._serialized_end=460
  _COLLABORATORDESCRIPTION._serialized_start=463
  _COLLABORATORDESCRIPTION._serialized_end=592
  _TASKDESCRIPTION._serialized_start=594
  _TASKDESCRIPTION._serialized_end=646
  _DOWNLOADSTATUS._serialized_start=648
  _DOWNLOADSTATUS._serialized_end=694
  _DOWNLOADSTATUSES._serialized_start=696
  _DOWNLOADSTATUSES._serialized_end=778
  _EXPERIMENTDESCRIPTION._serialized_start=781
  _EXPERIMENTDESCRIPTION._serialized_end=1025
# @@protoc_insertion_point(module_scope)
