# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: openfl/protocols/director.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2
from openfl.protocols import base_pb2 as openfl_dot_protocols_dot_base__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1fopenfl/protocols/director.proto\x12\x0fopenfl.director\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1egoogle/protobuf/duration.proto\x1a\x1bopenfl/protocols/base.proto\"\xab\x01\n\x0e\x43udaDeviceInfo\x12\r\n\x05index\x18\x01 \x01(\x04\x12\x14\n\x0cmemory_total\x18\x02 \x01(\x04\x12\x17\n\x0fmemory_utilized\x18\x03 \x01(\x04\x12\x1a\n\x12\x64\x65vice_utilization\x18\x04 \x01(\t\x12\x1b\n\x13\x63uda_driver_version\x18\x05 \x01(\t\x12\x14\n\x0c\x63uda_version\x18\x06 \x01(\t\x12\x0c\n\x04name\x18\x07 \x01(\t\"O\n\x08NodeInfo\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x35\n\x0c\x63uda_devices\x18\x02 \x03(\x0b\x32\x1f.openfl.director.CudaDeviceInfo\"\x93\x01\n\tShardInfo\x12,\n\tnode_info\x18\x01 \x01(\x0b\x32\x19.openfl.director.NodeInfo\x12\x19\n\x11shard_description\x18\x02 \x01(\t\x12\x11\n\tn_samples\x18\x03 \x01(\x04\x12\x14\n\x0csample_shape\x18\x04 \x03(\t\x12\x14\n\x0ctarget_shape\x18\x05 \x03(\t\"H\n\x16UpdateShardInfoRequest\x12.\n\nshard_info\x18\x01 \x01(\x0b\x32\x1a.openfl.director.ShardInfo\"+\n\x17UpdateShardInfoResponse\x12\x10\n\x08\x61\x63\x63\x65pted\x18\x01 \x01(\x08\"2\n\x15WaitExperimentRequest\x12\x19\n\x11\x63ollaborator_name\x18\x01 \x01(\t\"1\n\x16WaitExperimentResponse\x12\x17\n\x0f\x65xperiment_name\x18\x01 \x01(\t\"N\n\x18GetExperimentDataRequest\x12\x17\n\x0f\x65xperiment_name\x18\x01 \x01(\t\x12\x19\n\x11\x63ollaborator_name\x18\x02 \x01(\t\"/\n\x0e\x45xperimentData\x12\x0c\n\x04size\x18\x01 \x01(\r\x12\x0f\n\x07npbytes\x18\x02 \x01(\x0c\"~\n\x18UpdateEnvoyStatusRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x1d\n\x15is_experiment_running\x18\x02 \x01(\x08\x12\x35\n\x0c\x63uda_devices\x18\x03 \x03(\x0b\x32\x1f.openfl.director.CudaDeviceInfo\"S\n\x19UpdateEnvoyStatusResponse\x12\x36\n\x13health_check_period\x18\x01 \x01(\x0b\x32\x19.google.protobuf.Duration\"\x7f\n\x1aSetExperimentFailedRequest\x12\x17\n\x0f\x65xperiment_name\x18\x01 \x01(\t\x12\x19\n\x11\x63ollaborator_name\x18\x02 \x01(\t\x12\x12\n\nerror_code\x18\x03 \x01(\r\x12\x19\n\x11\x65rror_description\x18\x04 \x01(\t\"\x1d\n\x1bSetExperimentFailedResponse\"\x1b\n\x19GetExperimentsListRequest\"x\n\x12\x45xperimentListItem\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0e\n\x06status\x18\x02 \x01(\t\x12\x1c\n\x14\x63ollaborators_amount\x18\x03 \x01(\r\x12\x14\n\x0ctasks_amount\x18\x04 \x01(\r\x12\x10\n\x08progress\x18\x05 \x01(\x02\"V\n\x1aGetExperimentsListResponse\x12\x38\n\x0b\x65xperiments\x18\x01 \x03(\x0b\x32#.openfl.director.ExperimentListItem\"/\n\x1fGetExperimentDescriptionRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"N\n GetExperimentDescriptionResponse\x12*\n\nexperiment\x18\x01 \x01(\x0b\x32\x16.ExperimentDescription\"\x96\x01\n\x0e\x45xperimentInfo\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x1a\n\x12\x63ollaborator_names\x18\x02 \x03(\t\x12\x38\n\x0f\x65xperiment_data\x18\x03 \x01(\x0b\x32\x1f.openfl.director.ExperimentData\x12 \n\x0bmodel_proto\x18\x04 \x01(\x0b\x32\x0b.ModelProto\",\n\x18SetNewExperimentResponse\x12\x10\n\x08\x61\x63\x63\x65pted\x18\x01 \x01(\x08\"\xa5\x01\n\x16GetTrainedModelRequest\x12\x17\n\x0f\x65xperiment_name\x18\x02 \x01(\t\x12\x45\n\nmodel_type\x18\x03 \x01(\x0e\x32\x31.openfl.director.GetTrainedModelRequest.ModelType\"+\n\tModelType\x12\x0e\n\nBEST_MODEL\x10\x00\x12\x0e\n\nLAST_MODEL\x10\x01\"8\n\x14TrainedModelResponse\x12 \n\x0bmodel_proto\x18\x01 \x01(\x0b\x32\x0b.ModelProto\"\x17\n\x15GetDatasetInfoRequest\"H\n\x16GetDatasetInfoResponse\x12.\n\nshard_info\x18\x01 \x01(\x0b\x32\x1a.openfl.director.ShardInfo\"1\n\x16GetMetricStreamRequest\x12\x17\n\x0f\x65xperiment_name\x18\x01 \x01(\t\"}\n\x17GetMetricStreamResponse\x12\x15\n\rmetric_origin\x18\x01 \x01(\t\x12\x11\n\ttask_name\x18\x02 \x01(\t\x12\x13\n\x0bmetric_name\x18\x03 \x01(\t\x12\x14\n\x0cmetric_value\x18\x04 \x01(\x02\x12\r\n\x05round\x18\x05 \x01(\r\"2\n\x17RemoveExperimentRequest\x12\x17\n\x0f\x65xperiment_name\x18\x01 \x01(\t\"3\n\x18RemoveExperimentResponse\x12\x17\n\x0f\x61\x63knowledgement\x18\x01 \x01(\x08\"\xeb\x01\n\tEnvoyInfo\x12.\n\nshard_info\x18\x01 \x01(\x0b\x32\x1a.openfl.director.ShardInfo\x12\x11\n\tis_online\x18\x02 \x01(\x08\x12\x1d\n\x15is_experiment_running\x18\x03 \x01(\x08\x12\x30\n\x0clast_updated\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x31\n\x0evalid_duration\x18\x05 \x01(\x0b\x32\x19.google.protobuf.Duration\x12\x17\n\x0f\x65xperiment_name\x18\x06 \x01(\t\"\x12\n\x10GetEnvoysRequest\"D\n\x11GetEnvoysResponse\x12/\n\x0b\x65nvoy_infos\x18\x01 \x03(\x0b\x32\x1a.openfl.director.EnvoyInfo2\xf4\n\n\x08\x44irector\x12\x66\n\x0fUpdateShardInfo\x12\'.openfl.director.UpdateShardInfoRequest\x1a(.openfl.director.UpdateShardInfoResponse\"\x00\x12g\n\x0eWaitExperiment\x12&.openfl.director.WaitExperimentRequest\x1a\'.openfl.director.WaitExperimentResponse\"\x00(\x01\x30\x01\x12\x63\n\x11GetExperimentData\x12).openfl.director.GetExperimentDataRequest\x1a\x1f.openfl.director.ExperimentData\"\x00\x30\x01\x12l\n\x11UpdateEnvoyStatus\x12).openfl.director.UpdateEnvoyStatusRequest\x1a*.openfl.director.UpdateEnvoyStatusResponse\"\x00\x12r\n\x13SetExperimentFailed\x12+.openfl.director.SetExperimentFailedRequest\x1a,.openfl.director.SetExperimentFailedResponse\"\x00\x12\x81\x01\n\x18GetExperimentDescription\x12\x30.openfl.director.GetExperimentDescriptionRequest\x1a\x31.openfl.director.GetExperimentDescriptionResponse\"\x00\x12o\n\x12GetExperimentsList\x12*.openfl.director.GetExperimentsListRequest\x1a+.openfl.director.GetExperimentsListResponse\"\x00\x12\x62\n\x10SetNewExperiment\x12\x1f.openfl.director.ExperimentInfo\x1a).openfl.director.SetNewExperimentResponse\"\x00(\x01\x12\x63\n\x0eGetDatasetInfo\x12&.openfl.director.GetDatasetInfoRequest\x1a\'.openfl.director.GetDatasetInfoResponse\"\x00\x12\x63\n\x0fGetTrainedModel\x12\'.openfl.director.GetTrainedModelRequest\x1a%.openfl.director.TrainedModelResponse\"\x00\x12h\n\x0fGetMetricStream\x12\'.openfl.director.GetMetricStreamRequest\x1a(.openfl.director.GetMetricStreamResponse\"\x00\x30\x01\x12m\n\x14RemoveExperimentData\x12(.openfl.director.RemoveExperimentRequest\x1a).openfl.director.RemoveExperimentResponse\"\x00\x12T\n\tGetEnvoys\x12!.openfl.director.GetEnvoysRequest\x1a\".openfl.director.GetEnvoysResponse\"\x00\x62\x06proto3')



_CUDADEVICEINFO = DESCRIPTOR.message_types_by_name['CudaDeviceInfo']
_NODEINFO = DESCRIPTOR.message_types_by_name['NodeInfo']
_SHARDINFO = DESCRIPTOR.message_types_by_name['ShardInfo']
_UPDATESHARDINFOREQUEST = DESCRIPTOR.message_types_by_name['UpdateShardInfoRequest']
_UPDATESHARDINFORESPONSE = DESCRIPTOR.message_types_by_name['UpdateShardInfoResponse']
_WAITEXPERIMENTREQUEST = DESCRIPTOR.message_types_by_name['WaitExperimentRequest']
_WAITEXPERIMENTRESPONSE = DESCRIPTOR.message_types_by_name['WaitExperimentResponse']
_GETEXPERIMENTDATAREQUEST = DESCRIPTOR.message_types_by_name['GetExperimentDataRequest']
_EXPERIMENTDATA = DESCRIPTOR.message_types_by_name['ExperimentData']
_UPDATEENVOYSTATUSREQUEST = DESCRIPTOR.message_types_by_name['UpdateEnvoyStatusRequest']
_UPDATEENVOYSTATUSRESPONSE = DESCRIPTOR.message_types_by_name['UpdateEnvoyStatusResponse']
_SETEXPERIMENTFAILEDREQUEST = DESCRIPTOR.message_types_by_name['SetExperimentFailedRequest']
_SETEXPERIMENTFAILEDRESPONSE = DESCRIPTOR.message_types_by_name['SetExperimentFailedResponse']
_GETEXPERIMENTSLISTREQUEST = DESCRIPTOR.message_types_by_name['GetExperimentsListRequest']
_EXPERIMENTLISTITEM = DESCRIPTOR.message_types_by_name['ExperimentListItem']
_GETEXPERIMENTSLISTRESPONSE = DESCRIPTOR.message_types_by_name['GetExperimentsListResponse']
_GETEXPERIMENTDESCRIPTIONREQUEST = DESCRIPTOR.message_types_by_name['GetExperimentDescriptionRequest']
_GETEXPERIMENTDESCRIPTIONRESPONSE = DESCRIPTOR.message_types_by_name['GetExperimentDescriptionResponse']
_EXPERIMENTINFO = DESCRIPTOR.message_types_by_name['ExperimentInfo']
_SETNEWEXPERIMENTRESPONSE = DESCRIPTOR.message_types_by_name['SetNewExperimentResponse']
_GETTRAINEDMODELREQUEST = DESCRIPTOR.message_types_by_name['GetTrainedModelRequest']
_TRAINEDMODELRESPONSE = DESCRIPTOR.message_types_by_name['TrainedModelResponse']
_GETDATASETINFOREQUEST = DESCRIPTOR.message_types_by_name['GetDatasetInfoRequest']
_GETDATASETINFORESPONSE = DESCRIPTOR.message_types_by_name['GetDatasetInfoResponse']
_GETMETRICSTREAMREQUEST = DESCRIPTOR.message_types_by_name['GetMetricStreamRequest']
_GETMETRICSTREAMRESPONSE = DESCRIPTOR.message_types_by_name['GetMetricStreamResponse']
_REMOVEEXPERIMENTREQUEST = DESCRIPTOR.message_types_by_name['RemoveExperimentRequest']
_REMOVEEXPERIMENTRESPONSE = DESCRIPTOR.message_types_by_name['RemoveExperimentResponse']
_ENVOYINFO = DESCRIPTOR.message_types_by_name['EnvoyInfo']
_GETENVOYSREQUEST = DESCRIPTOR.message_types_by_name['GetEnvoysRequest']
_GETENVOYSRESPONSE = DESCRIPTOR.message_types_by_name['GetEnvoysResponse']
_GETTRAINEDMODELREQUEST_MODELTYPE = _GETTRAINEDMODELREQUEST.enum_types_by_name['ModelType']
CudaDeviceInfo = _reflection.GeneratedProtocolMessageType('CudaDeviceInfo', (_message.Message,), {
  'DESCRIPTOR' : _CUDADEVICEINFO,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.CudaDeviceInfo)
  })
_sym_db.RegisterMessage(CudaDeviceInfo)

NodeInfo = _reflection.GeneratedProtocolMessageType('NodeInfo', (_message.Message,), {
  'DESCRIPTOR' : _NODEINFO,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.NodeInfo)
  })
_sym_db.RegisterMessage(NodeInfo)

ShardInfo = _reflection.GeneratedProtocolMessageType('ShardInfo', (_message.Message,), {
  'DESCRIPTOR' : _SHARDINFO,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.ShardInfo)
  })
_sym_db.RegisterMessage(ShardInfo)

UpdateShardInfoRequest = _reflection.GeneratedProtocolMessageType('UpdateShardInfoRequest', (_message.Message,), {
  'DESCRIPTOR' : _UPDATESHARDINFOREQUEST,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.UpdateShardInfoRequest)
  })
_sym_db.RegisterMessage(UpdateShardInfoRequest)

UpdateShardInfoResponse = _reflection.GeneratedProtocolMessageType('UpdateShardInfoResponse', (_message.Message,), {
  'DESCRIPTOR' : _UPDATESHARDINFORESPONSE,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.UpdateShardInfoResponse)
  })
_sym_db.RegisterMessage(UpdateShardInfoResponse)

WaitExperimentRequest = _reflection.GeneratedProtocolMessageType('WaitExperimentRequest', (_message.Message,), {
  'DESCRIPTOR' : _WAITEXPERIMENTREQUEST,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.WaitExperimentRequest)
  })
_sym_db.RegisterMessage(WaitExperimentRequest)

WaitExperimentResponse = _reflection.GeneratedProtocolMessageType('WaitExperimentResponse', (_message.Message,), {
  'DESCRIPTOR' : _WAITEXPERIMENTRESPONSE,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.WaitExperimentResponse)
  })
_sym_db.RegisterMessage(WaitExperimentResponse)

GetExperimentDataRequest = _reflection.GeneratedProtocolMessageType('GetExperimentDataRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETEXPERIMENTDATAREQUEST,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.GetExperimentDataRequest)
  })
_sym_db.RegisterMessage(GetExperimentDataRequest)

ExperimentData = _reflection.GeneratedProtocolMessageType('ExperimentData', (_message.Message,), {
  'DESCRIPTOR' : _EXPERIMENTDATA,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.ExperimentData)
  })
_sym_db.RegisterMessage(ExperimentData)

UpdateEnvoyStatusRequest = _reflection.GeneratedProtocolMessageType('UpdateEnvoyStatusRequest', (_message.Message,), {
  'DESCRIPTOR' : _UPDATEENVOYSTATUSREQUEST,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.UpdateEnvoyStatusRequest)
  })
_sym_db.RegisterMessage(UpdateEnvoyStatusRequest)

UpdateEnvoyStatusResponse = _reflection.GeneratedProtocolMessageType('UpdateEnvoyStatusResponse', (_message.Message,), {
  'DESCRIPTOR' : _UPDATEENVOYSTATUSRESPONSE,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.UpdateEnvoyStatusResponse)
  })
_sym_db.RegisterMessage(UpdateEnvoyStatusResponse)

SetExperimentFailedRequest = _reflection.GeneratedProtocolMessageType('SetExperimentFailedRequest', (_message.Message,), {
  'DESCRIPTOR' : _SETEXPERIMENTFAILEDREQUEST,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.SetExperimentFailedRequest)
  })
_sym_db.RegisterMessage(SetExperimentFailedRequest)

SetExperimentFailedResponse = _reflection.GeneratedProtocolMessageType('SetExperimentFailedResponse', (_message.Message,), {
  'DESCRIPTOR' : _SETEXPERIMENTFAILEDRESPONSE,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.SetExperimentFailedResponse)
  })
_sym_db.RegisterMessage(SetExperimentFailedResponse)

GetExperimentsListRequest = _reflection.GeneratedProtocolMessageType('GetExperimentsListRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETEXPERIMENTSLISTREQUEST,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.GetExperimentsListRequest)
  })
_sym_db.RegisterMessage(GetExperimentsListRequest)

ExperimentListItem = _reflection.GeneratedProtocolMessageType('ExperimentListItem', (_message.Message,), {
  'DESCRIPTOR' : _EXPERIMENTLISTITEM,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.ExperimentListItem)
  })
_sym_db.RegisterMessage(ExperimentListItem)

GetExperimentsListResponse = _reflection.GeneratedProtocolMessageType('GetExperimentsListResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETEXPERIMENTSLISTRESPONSE,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.GetExperimentsListResponse)
  })
_sym_db.RegisterMessage(GetExperimentsListResponse)

GetExperimentDescriptionRequest = _reflection.GeneratedProtocolMessageType('GetExperimentDescriptionRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETEXPERIMENTDESCRIPTIONREQUEST,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.GetExperimentDescriptionRequest)
  })
_sym_db.RegisterMessage(GetExperimentDescriptionRequest)

GetExperimentDescriptionResponse = _reflection.GeneratedProtocolMessageType('GetExperimentDescriptionResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETEXPERIMENTDESCRIPTIONRESPONSE,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.GetExperimentDescriptionResponse)
  })
_sym_db.RegisterMessage(GetExperimentDescriptionResponse)

ExperimentInfo = _reflection.GeneratedProtocolMessageType('ExperimentInfo', (_message.Message,), {
  'DESCRIPTOR' : _EXPERIMENTINFO,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.ExperimentInfo)
  })
_sym_db.RegisterMessage(ExperimentInfo)

SetNewExperimentResponse = _reflection.GeneratedProtocolMessageType('SetNewExperimentResponse', (_message.Message,), {
  'DESCRIPTOR' : _SETNEWEXPERIMENTRESPONSE,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.SetNewExperimentResponse)
  })
_sym_db.RegisterMessage(SetNewExperimentResponse)

GetTrainedModelRequest = _reflection.GeneratedProtocolMessageType('GetTrainedModelRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETTRAINEDMODELREQUEST,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.GetTrainedModelRequest)
  })
_sym_db.RegisterMessage(GetTrainedModelRequest)

TrainedModelResponse = _reflection.GeneratedProtocolMessageType('TrainedModelResponse', (_message.Message,), {
  'DESCRIPTOR' : _TRAINEDMODELRESPONSE,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.TrainedModelResponse)
  })
_sym_db.RegisterMessage(TrainedModelResponse)

GetDatasetInfoRequest = _reflection.GeneratedProtocolMessageType('GetDatasetInfoRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETDATASETINFOREQUEST,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.GetDatasetInfoRequest)
  })
_sym_db.RegisterMessage(GetDatasetInfoRequest)

GetDatasetInfoResponse = _reflection.GeneratedProtocolMessageType('GetDatasetInfoResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETDATASETINFORESPONSE,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.GetDatasetInfoResponse)
  })
_sym_db.RegisterMessage(GetDatasetInfoResponse)

GetMetricStreamRequest = _reflection.GeneratedProtocolMessageType('GetMetricStreamRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETMETRICSTREAMREQUEST,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.GetMetricStreamRequest)
  })
_sym_db.RegisterMessage(GetMetricStreamRequest)

GetMetricStreamResponse = _reflection.GeneratedProtocolMessageType('GetMetricStreamResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETMETRICSTREAMRESPONSE,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.GetMetricStreamResponse)
  })
_sym_db.RegisterMessage(GetMetricStreamResponse)

RemoveExperimentRequest = _reflection.GeneratedProtocolMessageType('RemoveExperimentRequest', (_message.Message,), {
  'DESCRIPTOR' : _REMOVEEXPERIMENTREQUEST,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.RemoveExperimentRequest)
  })
_sym_db.RegisterMessage(RemoveExperimentRequest)

RemoveExperimentResponse = _reflection.GeneratedProtocolMessageType('RemoveExperimentResponse', (_message.Message,), {
  'DESCRIPTOR' : _REMOVEEXPERIMENTRESPONSE,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.RemoveExperimentResponse)
  })
_sym_db.RegisterMessage(RemoveExperimentResponse)

EnvoyInfo = _reflection.GeneratedProtocolMessageType('EnvoyInfo', (_message.Message,), {
  'DESCRIPTOR' : _ENVOYINFO,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.EnvoyInfo)
  })
_sym_db.RegisterMessage(EnvoyInfo)

GetEnvoysRequest = _reflection.GeneratedProtocolMessageType('GetEnvoysRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETENVOYSREQUEST,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.GetEnvoysRequest)
  })
_sym_db.RegisterMessage(GetEnvoysRequest)

GetEnvoysResponse = _reflection.GeneratedProtocolMessageType('GetEnvoysResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETENVOYSRESPONSE,
  '__module__' : 'openfl.protocols.director_pb2'
  # @@protoc_insertion_point(class_scope:openfl.director.GetEnvoysResponse)
  })
_sym_db.RegisterMessage(GetEnvoysResponse)

_DIRECTOR = DESCRIPTOR.services_by_name['Director']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _CUDADEVICEINFO._serialized_start=147
  _CUDADEVICEINFO._serialized_end=318
  _NODEINFO._serialized_start=320
  _NODEINFO._serialized_end=399
  _SHARDINFO._serialized_start=402
  _SHARDINFO._serialized_end=549
  _UPDATESHARDINFOREQUEST._serialized_start=551
  _UPDATESHARDINFOREQUEST._serialized_end=623
  _UPDATESHARDINFORESPONSE._serialized_start=625
  _UPDATESHARDINFORESPONSE._serialized_end=668
  _WAITEXPERIMENTREQUEST._serialized_start=670
  _WAITEXPERIMENTREQUEST._serialized_end=720
  _WAITEXPERIMENTRESPONSE._serialized_start=722
  _WAITEXPERIMENTRESPONSE._serialized_end=771
  _GETEXPERIMENTDATAREQUEST._serialized_start=773
  _GETEXPERIMENTDATAREQUEST._serialized_end=851
  _EXPERIMENTDATA._serialized_start=853
  _EXPERIMENTDATA._serialized_end=900
  _UPDATEENVOYSTATUSREQUEST._serialized_start=902
  _UPDATEENVOYSTATUSREQUEST._serialized_end=1028
  _UPDATEENVOYSTATUSRESPONSE._serialized_start=1030
  _UPDATEENVOYSTATUSRESPONSE._serialized_end=1113
  _SETEXPERIMENTFAILEDREQUEST._serialized_start=1115
  _SETEXPERIMENTFAILEDREQUEST._serialized_end=1242
  _SETEXPERIMENTFAILEDRESPONSE._serialized_start=1244
  _SETEXPERIMENTFAILEDRESPONSE._serialized_end=1273
  _GETEXPERIMENTSLISTREQUEST._serialized_start=1275
  _GETEXPERIMENTSLISTREQUEST._serialized_end=1302
  _EXPERIMENTLISTITEM._serialized_start=1304
  _EXPERIMENTLISTITEM._serialized_end=1424
  _GETEXPERIMENTSLISTRESPONSE._serialized_start=1426
  _GETEXPERIMENTSLISTRESPONSE._serialized_end=1512
  _GETEXPERIMENTDESCRIPTIONREQUEST._serialized_start=1514
  _GETEXPERIMENTDESCRIPTIONREQUEST._serialized_end=1561
  _GETEXPERIMENTDESCRIPTIONRESPONSE._serialized_start=1563
  _GETEXPERIMENTDESCRIPTIONRESPONSE._serialized_end=1641
  _EXPERIMENTINFO._serialized_start=1644
  _EXPERIMENTINFO._serialized_end=1794
  _SETNEWEXPERIMENTRESPONSE._serialized_start=1796
  _SETNEWEXPERIMENTRESPONSE._serialized_end=1840
  _GETTRAINEDMODELREQUEST._serialized_start=1843
  _GETTRAINEDMODELREQUEST._serialized_end=2008
  _GETTRAINEDMODELREQUEST_MODELTYPE._serialized_start=1965
  _GETTRAINEDMODELREQUEST_MODELTYPE._serialized_end=2008
  _TRAINEDMODELRESPONSE._serialized_start=2010
  _TRAINEDMODELRESPONSE._serialized_end=2066
  _GETDATASETINFOREQUEST._serialized_start=2068
  _GETDATASETINFOREQUEST._serialized_end=2091
  _GETDATASETINFORESPONSE._serialized_start=2093
  _GETDATASETINFORESPONSE._serialized_end=2165
  _GETMETRICSTREAMREQUEST._serialized_start=2167
  _GETMETRICSTREAMREQUEST._serialized_end=2216
  _GETMETRICSTREAMRESPONSE._serialized_start=2218
  _GETMETRICSTREAMRESPONSE._serialized_end=2343
  _REMOVEEXPERIMENTREQUEST._serialized_start=2345
  _REMOVEEXPERIMENTREQUEST._serialized_end=2395
  _REMOVEEXPERIMENTRESPONSE._serialized_start=2397
  _REMOVEEXPERIMENTRESPONSE._serialized_end=2448
  _ENVOYINFO._serialized_start=2451
  _ENVOYINFO._serialized_end=2686
  _GETENVOYSREQUEST._serialized_start=2688
  _GETENVOYSREQUEST._serialized_end=2706
  _GETENVOYSRESPONSE._serialized_start=2708
  _GETENVOYSRESPONSE._serialized_end=2776
  _DIRECTOR._serialized_start=2779
  _DIRECTOR._serialized_end=4175
# @@protoc_insertion_point(module_scope)
