# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Aggregation functions experimental package."""

from .privileged_aggregation import PrivilegedAggregationFunction

__all__ = ['PrivilegedAggregationFunction',]
