# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: openfl/protocols/aggregator.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from openfl.protocols import base_pb2 as openfl_dot_protocols_dot_base__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n!openfl/protocols/aggregator.proto\x12\x11openfl.aggregator\x1a\x1bopenfl/protocols/base.proto\"o\n\rMessageHeader\x12\x0e\n\x06sender\x18\x01 \x01(\t\x12\x10\n\x08receiver\x18\x02 \x01(\t\x12\x17\n\x0f\x66\x65\x64\x65ration_uuid\x18\x03 \x01(\t\x12#\n\x1bsingle_col_cert_common_name\x18\x04 \x01(\t\"C\n\x0fGetTasksRequest\x12\x30\n\x06header\x18\x01 \x01(\x0b\x32 .openfl.aggregator.MessageHeader\"S\n\x04Task\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x15\n\rfunction_name\x18\x02 \x01(\t\x12\x11\n\ttask_type\x18\x03 \x01(\t\x12\x13\n\x0b\x61pply_local\x18\x04 \x01(\x08\"\xa4\x01\n\x10GetTasksResponse\x12\x30\n\x06header\x18\x01 \x01(\x0b\x32 .openfl.aggregator.MessageHeader\x12\x14\n\x0cround_number\x18\x02 \x01(\x05\x12&\n\x05tasks\x18\x03 \x03(\x0b\x32\x17.openfl.aggregator.Task\x12\x12\n\nsleep_time\x18\x04 \x01(\x05\x12\x0c\n\x04quit\x18\x05 \x01(\x08\"\xbb\x01\n\x10GetTensorRequest\x12\x30\n\x06header\x18\x01 \x01(\x0b\x32 .openfl.aggregator.MessageHeader\x12\x13\n\x0btensor_name\x18\x02 \x01(\t\x12\x14\n\x0cround_number\x18\x03 \x01(\x05\x12\x0e\n\x06report\x18\x04 \x01(\x08\x12\x0c\n\x04tags\x18\x05 \x03(\t\x12\x18\n\x10require_lossless\x18\x06 \x01(\x08\x12\x12\n\naggregated\x18\x07 \x01(\x08\"y\n\x11GetTensorResponse\x12\x30\n\x06header\x18\x01 \x01(\x0b\x32 .openfl.aggregator.MessageHeader\x12\x14\n\x0cround_number\x18\x02 \x01(\x05\x12\x1c\n\x06tensor\x18\x03 \x01(\x0b\x32\x0c.NamedTensor\"\x9a\x01\n\x0bTaskResults\x12\x30\n\x06header\x18\x01 \x01(\x0b\x32 .openfl.aggregator.MessageHeader\x12\x14\n\x0cround_number\x18\x02 \x01(\x05\x12\x11\n\ttask_name\x18\x03 \x01(\t\x12\x11\n\tdata_size\x18\x04 \x01(\x05\x12\x1d\n\x07tensors\x18\x05 \x03(\x0b\x32\x0c.NamedTensor\"P\n\x1cSendLocalTaskResultsResponse\x12\x30\n\x06header\x18\x01 \x01(\x0b\x32 .openfl.aggregator.MessageHeader\"1\n\x16GetMetricStreamRequest\x12\x17\n\x0f\x65xperiment_name\x18\x01 \x01(\t\"}\n\x17GetMetricStreamResponse\x12\x15\n\rmetric_origin\x18\x01 \x01(\t\x12\x11\n\ttask_name\x18\x02 \x01(\t\x12\x13\n\x0bmetric_name\x18\x03 \x01(\t\x12\x14\n\x0cmetric_value\x18\x04 \x01(\x02\x12\r\n\x05round\x18\x05 \x01(\r\"\xa7\x01\n\x16GetTrainedModelRequest\x12\x17\n\x0f\x65xperiment_name\x18\x02 \x01(\t\x12G\n\nmodel_type\x18\x03 \x01(\x0e\x32\x33.openfl.aggregator.GetTrainedModelRequest.ModelType\"+\n\tModelType\x12\x0e\n\nBEST_MODEL\x10\x00\x12\x0e\n\nLAST_MODEL\x10\x01\"8\n\x14TrainedModelResponse\x12 \n\x0bmodel_proto\x18\x01 \x01(\x0b\x32\x0b.ModelProto\"/\n\x1fGetExperimentDescriptionRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"N\n GetExperimentDescriptionResponse\x12*\n\nexperiment\x18\x01 \x01(\x0b\x32\x16.ExperimentDescription\"i\n\x0cSynchRequest\x12\x30\n\x06header\x18\x01 \x01(\x0b\x32 .openfl.aggregator.MessageHeader\x12\x11\n\ttask_name\x18\x02 \x01(\t\x12\x14\n\x0cround_number\x18\x03 \x01(\x05\"W\n\rSynchResponse\x12\x30\n\x06header\x18\x01 \x01(\x0b\x32 .openfl.aggregator.MessageHeader\x12\x14\n\x0cis_completed\x18\x02 \x01(\x08\x32\xc7\x05\n\nAggregator\x12U\n\x08GetTasks\x12\".openfl.aggregator.GetTasksRequest\x1a#.openfl.aggregator.GetTasksResponse\"\x00\x12O\n\x08GetSynch\x12\x1f.openfl.aggregator.SynchRequest\x1a .openfl.aggregator.SynchResponse\"\x00\x12X\n\tGetTensor\x12#.openfl.aggregator.GetTensorRequest\x1a$.openfl.aggregator.GetTensorResponse\"\x00\x12X\n\x14SendLocalTaskResults\x12\x0b.DataStream\x1a/.openfl.aggregator.SendLocalTaskResultsResponse\"\x00(\x01\x12l\n\x0fGetMetricStream\x12).openfl.aggregator.GetMetricStreamRequest\x1a*.openfl.aggregator.GetMetricStreamResponse\"\x00\x30\x01\x12g\n\x0fGetTrainedModel\x12).openfl.aggregator.GetTrainedModelRequest\x1a\'.openfl.aggregator.TrainedModelResponse\"\x00\x12\x85\x01\n\x18GetExperimentDescription\x12\x32.openfl.aggregator.GetExperimentDescriptionRequest\x1a\x33.openfl.aggregator.GetExperimentDescriptionResponse\"\x00\x62\x06proto3')



_MESSAGEHEADER = DESCRIPTOR.message_types_by_name['MessageHeader']
_GETTASKSREQUEST = DESCRIPTOR.message_types_by_name['GetTasksRequest']
_TASK = DESCRIPTOR.message_types_by_name['Task']
_GETTASKSRESPONSE = DESCRIPTOR.message_types_by_name['GetTasksResponse']
_GETTENSORREQUEST = DESCRIPTOR.message_types_by_name['GetTensorRequest']
_GETTENSORRESPONSE = DESCRIPTOR.message_types_by_name['GetTensorResponse']
_TASKRESULTS = DESCRIPTOR.message_types_by_name['TaskResults']
_SENDLOCALTASKRESULTSRESPONSE = DESCRIPTOR.message_types_by_name['SendLocalTaskResultsResponse']
_GETMETRICSTREAMREQUEST = DESCRIPTOR.message_types_by_name['GetMetricStreamRequest']
_GETMETRICSTREAMRESPONSE = DESCRIPTOR.message_types_by_name['GetMetricStreamResponse']
_GETTRAINEDMODELREQUEST = DESCRIPTOR.message_types_by_name['GetTrainedModelRequest']
_TRAINEDMODELRESPONSE = DESCRIPTOR.message_types_by_name['TrainedModelResponse']
_GETEXPERIMENTDESCRIPTIONREQUEST = DESCRIPTOR.message_types_by_name['GetExperimentDescriptionRequest']
_GETEXPERIMENTDESCRIPTIONRESPONSE = DESCRIPTOR.message_types_by_name['GetExperimentDescriptionResponse']
_SYNCHREQUEST = DESCRIPTOR.message_types_by_name['SynchRequest']
_SYNCHRESPONSE = DESCRIPTOR.message_types_by_name['SynchResponse']
_GETTRAINEDMODELREQUEST_MODELTYPE = _GETTRAINEDMODELREQUEST.enum_types_by_name['ModelType']
MessageHeader = _reflection.GeneratedProtocolMessageType('MessageHeader', (_message.Message,), {
  'DESCRIPTOR' : _MESSAGEHEADER,
  '__module__' : 'openfl.protocols.aggregator_pb2'
  # @@protoc_insertion_point(class_scope:openfl.aggregator.MessageHeader)
  })
_sym_db.RegisterMessage(MessageHeader)

GetTasksRequest = _reflection.GeneratedProtocolMessageType('GetTasksRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETTASKSREQUEST,
  '__module__' : 'openfl.protocols.aggregator_pb2'
  # @@protoc_insertion_point(class_scope:openfl.aggregator.GetTasksRequest)
  })
_sym_db.RegisterMessage(GetTasksRequest)

Task = _reflection.GeneratedProtocolMessageType('Task', (_message.Message,), {
  'DESCRIPTOR' : _TASK,
  '__module__' : 'openfl.protocols.aggregator_pb2'
  # @@protoc_insertion_point(class_scope:openfl.aggregator.Task)
  })
_sym_db.RegisterMessage(Task)

GetTasksResponse = _reflection.GeneratedProtocolMessageType('GetTasksResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETTASKSRESPONSE,
  '__module__' : 'openfl.protocols.aggregator_pb2'
  # @@protoc_insertion_point(class_scope:openfl.aggregator.GetTasksResponse)
  })
_sym_db.RegisterMessage(GetTasksResponse)

GetTensorRequest = _reflection.GeneratedProtocolMessageType('GetTensorRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETTENSORREQUEST,
  '__module__' : 'openfl.protocols.aggregator_pb2'
  # @@protoc_insertion_point(class_scope:openfl.aggregator.GetTensorRequest)
  })
_sym_db.RegisterMessage(GetTensorRequest)

GetTensorResponse = _reflection.GeneratedProtocolMessageType('GetTensorResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETTENSORRESPONSE,
  '__module__' : 'openfl.protocols.aggregator_pb2'
  # @@protoc_insertion_point(class_scope:openfl.aggregator.GetTensorResponse)
  })
_sym_db.RegisterMessage(GetTensorResponse)

TaskResults = _reflection.GeneratedProtocolMessageType('TaskResults', (_message.Message,), {
  'DESCRIPTOR' : _TASKRESULTS,
  '__module__' : 'openfl.protocols.aggregator_pb2'
  # @@protoc_insertion_point(class_scope:openfl.aggregator.TaskResults)
  })
_sym_db.RegisterMessage(TaskResults)

SendLocalTaskResultsResponse = _reflection.GeneratedProtocolMessageType('SendLocalTaskResultsResponse', (_message.Message,), {
  'DESCRIPTOR' : _SENDLOCALTASKRESULTSRESPONSE,
  '__module__' : 'openfl.protocols.aggregator_pb2'
  # @@protoc_insertion_point(class_scope:openfl.aggregator.SendLocalTaskResultsResponse)
  })
_sym_db.RegisterMessage(SendLocalTaskResultsResponse)

GetMetricStreamRequest = _reflection.GeneratedProtocolMessageType('GetMetricStreamRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETMETRICSTREAMREQUEST,
  '__module__' : 'openfl.protocols.aggregator_pb2'
  # @@protoc_insertion_point(class_scope:openfl.aggregator.GetMetricStreamRequest)
  })
_sym_db.RegisterMessage(GetMetricStreamRequest)

GetMetricStreamResponse = _reflection.GeneratedProtocolMessageType('GetMetricStreamResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETMETRICSTREAMRESPONSE,
  '__module__' : 'openfl.protocols.aggregator_pb2'
  # @@protoc_insertion_point(class_scope:openfl.aggregator.GetMetricStreamResponse)
  })
_sym_db.RegisterMessage(GetMetricStreamResponse)

GetTrainedModelRequest = _reflection.GeneratedProtocolMessageType('GetTrainedModelRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETTRAINEDMODELREQUEST,
  '__module__' : 'openfl.protocols.aggregator_pb2'
  # @@protoc_insertion_point(class_scope:openfl.aggregator.GetTrainedModelRequest)
  })
_sym_db.RegisterMessage(GetTrainedModelRequest)

TrainedModelResponse = _reflection.GeneratedProtocolMessageType('TrainedModelResponse', (_message.Message,), {
  'DESCRIPTOR' : _TRAINEDMODELRESPONSE,
  '__module__' : 'openfl.protocols.aggregator_pb2'
  # @@protoc_insertion_point(class_scope:openfl.aggregator.TrainedModelResponse)
  })
_sym_db.RegisterMessage(TrainedModelResponse)

GetExperimentDescriptionRequest = _reflection.GeneratedProtocolMessageType('GetExperimentDescriptionRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETEXPERIMENTDESCRIPTIONREQUEST,
  '__module__' : 'openfl.protocols.aggregator_pb2'
  # @@protoc_insertion_point(class_scope:openfl.aggregator.GetExperimentDescriptionRequest)
  })
_sym_db.RegisterMessage(GetExperimentDescriptionRequest)

GetExperimentDescriptionResponse = _reflection.GeneratedProtocolMessageType('GetExperimentDescriptionResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETEXPERIMENTDESCRIPTIONRESPONSE,
  '__module__' : 'openfl.protocols.aggregator_pb2'
  # @@protoc_insertion_point(class_scope:openfl.aggregator.GetExperimentDescriptionResponse)
  })
_sym_db.RegisterMessage(GetExperimentDescriptionResponse)

SynchRequest = _reflection.GeneratedProtocolMessageType('SynchRequest', (_message.Message,), {
  'DESCRIPTOR' : _SYNCHREQUEST,
  '__module__' : 'openfl.protocols.aggregator_pb2'
  # @@protoc_insertion_point(class_scope:openfl.aggregator.SynchRequest)
  })
_sym_db.RegisterMessage(SynchRequest)

SynchResponse = _reflection.GeneratedProtocolMessageType('SynchResponse', (_message.Message,), {
  'DESCRIPTOR' : _SYNCHRESPONSE,
  '__module__' : 'openfl.protocols.aggregator_pb2'
  # @@protoc_insertion_point(class_scope:openfl.aggregator.SynchResponse)
  })
_sym_db.RegisterMessage(SynchResponse)

_AGGREGATOR = DESCRIPTOR.services_by_name['Aggregator']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _MESSAGEHEADER._serialized_start=85
  _MESSAGEHEADER._serialized_end=196
  _GETTASKSREQUEST._serialized_start=198
  _GETTASKSREQUEST._serialized_end=265
  _TASK._serialized_start=267
  _TASK._serialized_end=350
  _GETTASKSRESPONSE._serialized_start=353
  _GETTASKSRESPONSE._serialized_end=517
  _GETTENSORREQUEST._serialized_start=520
  _GETTENSORREQUEST._serialized_end=707
  _GETTENSORRESPONSE._serialized_start=709
  _GETTENSORRESPONSE._serialized_end=830
  _TASKRESULTS._serialized_start=833
  _TASKRESULTS._serialized_end=987
  _SENDLOCALTASKRESULTSRESPONSE._serialized_start=989
  _SENDLOCALTASKRESULTSRESPONSE._serialized_end=1069
  _GETMETRICSTREAMREQUEST._serialized_start=1071
  _GETMETRICSTREAMREQUEST._serialized_end=1120
  _GETMETRICSTREAMRESPONSE._serialized_start=1122
  _GETMETRICSTREAMRESPONSE._serialized_end=1247
  _GETTRAINEDMODELREQUEST._serialized_start=1250
  _GETTRAINEDMODELREQUEST._serialized_end=1417
  _GETTRAINEDMODELREQUEST_MODELTYPE._serialized_start=1374
  _GETTRAINEDMODELREQUEST_MODELTYPE._serialized_end=1417
  _TRAINEDMODELRESPONSE._serialized_start=1419
  _TRAINEDMODELRESPONSE._serialized_end=1475
  _GETEXPERIMENTDESCRIPTIONREQUEST._serialized_start=1477
  _GETEXPERIMENTDESCRIPTIONREQUEST._serialized_end=1524
  _GETEXPERIMENTDESCRIPTIONRESPONSE._serialized_start=1526
  _GETEXPERIMENTDESCRIPTIONRESPONSE._serialized_end=1604
  _SYNCHREQUEST._serialized_start=1606
  _SYNCHREQUEST._serialized_end=1711
  _SYNCHRESPONSE._serialized_start=1713
  _SYNCHRESPONSE._serialized_end=1800
  _AGGREGATOR._serialized_start=1803
  _AGGREGATOR._serialized_end=2514
# @@protoc_insertion_point(module_scope)
